use icinga_api::api::object_type::ObjectType;
use icinga_api::{api, icinga};
use std::convert::TryFrom;
use std::sync::Arc;
use tracing::{debug, error};

pub async fn do_actions(
    service: &Arc<icinga::Service>,
    client: &Arc<reqwest::Client>,
    parse_result: i2api_utils::parsing::ParseResult,
    author: &str,
) -> Result<(), icinga_api::utils::Error> {
    use i2api_utils::parsing;

    let service_name = service.name();
    let service = service.clone();
    let client = client.clone();
    let action_handle = match parse_result {
        parsing::ParseResult::Highlight | parsing::ParseResult::Help => unimplemented!(),
        parsing::ParseResult::Ack(p_res) => acknowledge(p_res, service, client, author)?,
        parsing::ParseResult::AckExp(p_res) => acknowledge_exp(p_res, service, client, author)?,
        parsing::ParseResult::Dt(p_res) => downtime(p_res, service, client, author)?,
        parsing::ParseResult::Com(p_res) => comment(p_res, service, client, author)?,
        parsing::ParseResult::Notif(active) => set_notification(active, service, client),
        parsing::ParseResult::Rm(p_res) => {
            let join_handle = remove(p_res, service, client, author)?;
            if let Ok(handles) = join_handle.await {
                for res in handles {
                    match res {
                        Ok(action_result) => {
                            if action_result.results.is_empty() {
                                println!("Nothing to do -- {service_name}");
                            }

                            for res in action_result.results {
                                println!("{} -- {service_name}", res.status);
                            }
                        }
                        Err(err) => error!("{} -- {}", err, service_name),
                    }
                }
            };
            return Ok(());
        }
        parsing::ParseResult::Unknown => return Ok(()),
    };
    let action_result = action_handle.await;
    if let Ok(action_result) = action_result {
        match action_result {
            Ok(action_result) => {
                if action_result.results.is_empty() {
                    println!("Nothing to do -- {service_name}");
                }

                for res in action_result.results {
                    println!("{} -- {service_name}", res.status);
                }
            }
            Err(err) => error!("{} -- {}", err, service_name),
        }
    }
    Ok(())
}

type ActionResult = Result<
    tokio::task::JoinHandle<Result<icinga_api::api::ActionReply, reqwest::Error>>,
    icinga_api::utils::Error,
>;

fn acknowledge(
    p_res: i2api_utils::parsing::AckParsed,
    service: Arc<icinga::Service>,
    client: Arc<reqwest::Client>,
    author: &str,
) -> ActionResult {
    let filter = icinga_api::utils::NonEmptyStr::try_from(p_res.filter)?;
    let action = api::AcknowledgeProblemActionBuilder::new(filter, p_res.a_type)
        .unwrap()
        .filter_vars(p_res.filter_vars)
        .comment(p_res.comment)
        .author(author)
        .build();
    debug!("Ack: {:?}", action);
    Ok(tokio::spawn(async move {
        service.acknowledge_problem(&client, &action).await
    }))
}

fn acknowledge_exp(
    p_res: i2api_utils::parsing::AckExpParsed,
    service: Arc<icinga::Service>,
    client: Arc<reqwest::Client>,
    author: &str,
) -> ActionResult {
    let filter = icinga_api::utils::NonEmptyStr::try_from(p_res.filter)?;
    let action = api::AcknowledgeProblemActionBuilder::new(filter, p_res.a_type)
        .unwrap()
        .filter_vars(p_res.filter_vars)
        .comment(p_res.comment)
        .expiry(p_res.end_time)
        .author(author)
        .build();
    debug!("Ackexp: {:?}", action);
    Ok(tokio::spawn(async move {
        service.acknowledge_problem(&client, &action).await
    }))
}

fn downtime(
    p_res: i2api_utils::parsing::DtParsed,
    service: Arc<icinga::Service>,
    client: Arc<reqwest::Client>,
    author: &str,
) -> ActionResult {
    let filter = icinga_api::utils::NonEmptyStr::try_from(p_res.filter)?;
    let comment = icinga_api::utils::NonEmptyStr::try_from(p_res.comment)?;
    let author = icinga_api::utils::NonEmptyStr::try_from(author).unwrap();
    let action = api::ScheduleDowntimeActionBuilder::new(
        author,
        comment,
        p_res.end_time,
        filter,
        p_res.a_type,
    )?
    .filter_vars(p_res.filter_vars)
    .all_services(matches!(p_res.a_type, ObjectType::Host))
    .build();
    debug!("Downtime: {:?}", action);
    Ok(tokio::spawn(async move {
        service.schedule_downtime(&client, &action).await
    }))
}

fn comment(
    p_res: i2api_utils::parsing::ComParsed,
    service: Arc<icinga::Service>,
    client: Arc<reqwest::Client>,
    author: &str,
) -> ActionResult {
    let filter = icinga_api::utils::NonEmptyStr::try_from(p_res.filter)?;
    let action = api::AddCommentActionBuilder::new(filter, p_res.a_type)?
        .filter_vars(p_res.filter_vars)
        .comment(p_res.comment)
        .author(author)
        .build();
    debug!("Comment: {:?}", action);
    Ok(tokio::spawn(async move {
        service.add_comment(&client, &action).await
    }))
}

fn set_notification(
    active: bool,
    service: Arc<icinga::Service>,
    client: Arc<reqwest::Client>,
) -> tokio::task::JoinHandle<Result<icinga_api::api::ActionReply, reqwest::Error>> {
    debug!("Notification: {:?}", active);
    tokio::spawn(async move { service.set_notification(&client, active).await })
}

fn remove(
    p_res: i2api_utils::parsing::RmParsed,
    service: Arc<icinga::Service>,
    client: Arc<reqwest::Client>,
    author: &str,
) -> Result<
    futures::future::TryJoinAll<
        tokio::task::JoinHandle<Result<icinga_api::api::ActionReply, reqwest::Error>>,
    >,
    icinga_api::utils::Error,
> {
    let filter = icinga_api::utils::NonEmptyStr::try_from(p_res.filter)?;
    let dt_type = match p_res.a_type {
        ObjectType::Service => p_res.a_type,
        ObjectType::Host => ObjectType::Downtime,
        _ => unreachable!(),
    };
    let dt_action = api::RemoveDowntimeActionBuilder::new(filter.clone(), dt_type)?
        .filter_vars(p_res.filter_vars.clone())
        .author(author)
        .build();
    let dt_handle = {
        let service = service.clone();
        let client = client.clone();
        tokio::spawn(async move { service.remove_downtime(&client, &dt_action).await })
    };
    let ack_action1 = api::RemoveAcknowledgementActionBuilder::new(filter.clone(), p_res.a_type)?
        .filter_vars(p_res.filter_vars.clone())
        .author(author)
        .build();
    let ack_handle1 = {
        let service = service.clone();
        let client = client.clone();
        tokio::spawn(async move { service.remove_acknowledgement(&client, &ack_action1).await })
    };
    let join_handle = match p_res.a_type {
        ObjectType::Service => futures::future::try_join_all([dt_handle, ack_handle1]),
        ObjectType::Host => {
            let ack_action2 =
                api::RemoveAcknowledgementActionBuilder::new(filter, ObjectType::Service)?
                    .filter_vars(p_res.filter_vars)
                    .author(author)
                    .build();
            let ack_handle2 =
                tokio::spawn(
                    async move { service.remove_acknowledgement(&client, &ack_action2).await },
                );
            futures::future::try_join_all([dt_handle, ack_handle1, ack_handle2])
        }
        _ => unreachable!(),
    };
    Ok(join_handle)
}
