#![deny(unsafe_code)]

extern crate i2api_utils;
extern crate icinga_api;
extern crate toml;

use std::collections::HashMap;
use std::convert::{TryFrom, TryInto};
use std::fmt::Debug;
use std::fs;
use std::path::PathBuf;
use std::sync::Arc;

use futures_util::future::join_all;
use hyper::client::connect::Connect;
use hyper_rustls::HttpsConnectorBuilder;
use i2api_utils::conf;
use i2api_utils::parsing::FilterVars;
use icinga_api::icinga;
use rustls::ClientConfig;

mod cli;
mod tls;

fn main() -> Result<(), Box<dyn std::error::Error>> {
    let args = cli::build().get_matches();

    let conf_file = fs::read_to_string(args.get_one::<std::path::PathBuf>("config").unwrap())?;
    let config: conf::Config = toml::from_str(&conf_file)?;

    let client = {
        let client_config = ClientConfig::builder()
            .with_safe_defaults()
            .with_custom_certificate_verifier(Arc::new(tls::NoVerifier))
            .with_no_client_auth();
        let connector = HttpsConnectorBuilder::new()
            .with_tls_config(client_config)
            .https_or_http()
            .enable_http1()
            .build();
        hyper::Client::builder()
            .retry_canceled_requests(true)
            .build(connector)
    };

    let author = args
        .get_one::<String>("author")
        .map(ToString::to_string)
        .or(config.author)
        .or_else(|| std::env::var("USER").ok())
        .expect("Author must be defined");

    let icingas = config
        .hosts
        .into_iter()
        .map(|(srv, conf)| {
            let password = password_from_config(&conf);

            icinga::ServiceBuilder::new(client.clone())
                .url(conf.url)
                .auth(conf.username, password)
                .name(srv)
                .build()
                .unwrap()
        })
        .map(Arc::new);

    let rt = tokio::runtime::Builder::new_multi_thread()
        .enable_all()
        .build()?;

    match args.subcommand() {
        Some(("ack", args)) => ack(&rt, &author, icingas, args),
        Some(("ackexp", args)) => ackexp(&rt, &author, icingas, args)?,
        Some(("com", args)) => com(&rt, &author, icingas, args),
        Some(("dt", args)) => dt(&rt, &author, icingas, args)?,
        Some(("rm", args)) => rm(&rt, &author, icingas, args),
        _ => unreachable!(),
    };
    Ok(())
}

fn password_from_config(conf: &conf::CService) -> String {
    conf.password_file
        .as_ref()
        .and_then(|password_file| {
            let password_file = PathBuf::from(password_file);
            std::fs::read_to_string(password_file).ok()
        })
        .or_else(|| conf.password.clone())
        .expect("A password has to be provided, either through `password` or a file (`password_file`)")
}

fn parse_target(args: &clap::ArgMatches) -> i2api_utils::parsing::Target {
    let mut filter_vars = FilterVars::default();
    let mut filter = "regex(HostName, host.name)".to_string();
    let mut a_type = icinga_api::api::object_type::ObjectType::Host;

    filter_vars.host = format!("^{}$", args.get_one::<String>("host").unwrap());
    if let Some(service) = args.get_one::<String>("service") {
        filter_vars.service = Some(format!("^{service}$"));
        filter.push_str(" && regex(ServiceName, service.name)");
        a_type = icinga_api::api::object_type::ObjectType::Service;
    }
    if let Some(hoster) = args.get_one::<String>("hoster") {
        filter_vars.hoster = Some(format!("^{hoster}$"));
        filter.push_str(" && regex(Hoster, host.vars.computer_hoster)");
    }

    i2api_utils::parsing::Target {
        a_type,
        filter: filter.to_string(),
        filter_vars,
    }
}

fn ack<Connector>(
    rt: &tokio::runtime::Runtime,
    author: &str,
    icingas: impl Iterator<Item = Arc<icinga::Service<Connector>>>,
    ack_args: &clap::ArgMatches,
) where
    Connector: Connect + Clone + Send + Sync + Debug + 'static,
{
    let comment: Vec<_> = ack_args
        .get_many::<String>("comment")
        .unwrap()
        .map(String::as_str)
        .flat_map(str::split_whitespace)
        .collect();
    let comment = i2api_utils::parsing::parse_comment(&comment);

    let target = parse_target(ack_args);

    let filter = icinga_api::utils::NonEmptyStr::try_from(target.filter).unwrap();

    let action = icinga_api::api::AcknowledgeProblemAction::builder(
        filter,
        target.a_type.try_into().unwrap(),
    )
    .filter_vars(target.filter_vars.into())
    .sticky(ack_args.get_flag("sticky"))
    .comment(comment)
    .author(author)
    .build();
    let action = Arc::new(action);

    rt.block_on(join_all(icingas.map(|service| {
        let action = action.clone();
        rt.spawn(async move {
            service.acknowledge_problem(&action).await.unwrap();
        })
    })));
}

fn ackexp<Connector>(
    rt: &tokio::runtime::Runtime,
    author: &str,
    icingas: impl Iterator<Item = Arc<icinga::Service<Connector>>>,
    ack_args: &clap::ArgMatches,
) -> Result<(), i2api_utils::parsing::ParseEndTimeError>
where
    Connector: Connect + Clone + Send + Sync + Debug + 'static,
{
    let action = {
        let end_time = time::OffsetDateTime::now_utc()
            + i2api_utils::parsing::parse_end_time(
                ack_args.get_one::<String>("end_time").unwrap(),
            )?;

        let comment: Vec<_> = ack_args
            .get_many::<String>("comment")
            .unwrap()
            .map(String::as_str)
            .flat_map(str::split_whitespace)
            .collect();
        let comment = i2api_utils::parsing::parse_comment(&comment);

        let target = parse_target(ack_args);

        let filter = icinga_api::utils::NonEmptyStr::try_from(target.filter).unwrap();

        let action = icinga_api::api::AcknowledgeProblemAction::builder(
            filter,
            target.a_type.try_into().unwrap(),
        )
        .filter_vars(target.filter_vars.into())
        .sticky(ack_args.get_flag("sticky"))
        .comment(comment)
        .author(author)
        .expiry(end_time)
        .build();
        Arc::new(action)
    };

    rt.block_on(join_all(icingas.map(|service| {
        let action = action.clone();
        rt.spawn(async move {
            service.acknowledge_problem(&action).await.unwrap();
        })
    })));

    Ok(())
}

fn com<Connector>(
    rt: &tokio::runtime::Runtime,
    author: &str,
    icingas: impl Iterator<Item = Arc<icinga::Service<Connector>>>,
    com_args: &clap::ArgMatches,
) where
    Connector: Connect + Clone + Send + Sync + Debug + 'static,
{
    let action = {
        let comment: Vec<_> = com_args
            .get_many::<String>("comment")
            .unwrap()
            .map(String::as_str)
            .flat_map(str::split_whitespace)
            .collect();
        let comment = i2api_utils::parsing::parse_comment(&comment);

        let target = parse_target(com_args);

        let filter = icinga_api::utils::NonEmptyStr::try_from(target.filter).unwrap();

        let action =
            icinga_api::api::AddCommentAction::builder(filter, target.a_type.try_into().unwrap())
                .filter_vars(target.filter_vars.into())
                .comment(comment)
                .author(author)
                .build();
        Arc::new(action)
    };

    rt.block_on(join_all(icingas.map(|service| {
        let action = action.clone();
        rt.spawn(async move {
            service.add_comment(&action).await.unwrap();
        })
    })));
}

fn dt<Connector>(
    rt: &tokio::runtime::Runtime,
    author: &str,
    icingas: impl Iterator<Item = Arc<icinga::Service<Connector>>>,
    dt_args: &clap::ArgMatches,
) -> Result<(), i2api_utils::parsing::ParseEndTimeError>
where
    Connector: Connect + Clone + Send + Sync + Debug + 'static,
{
    let action = {
        let start_time = dt_args
            .get_one::<String>("start_time")
            .map(|st| {
                time::OffsetDateTime::parse(st, &time::format_description::well_known::Rfc3339)
            })
            .transpose()
            .unwrap()
            .unwrap_or_else(time::OffsetDateTime::now_utc);

        let end_time = start_time
            + i2api_utils::parsing::parse_end_time(dt_args.get_one::<String>("end_time").unwrap())?;

        let comment: Vec<_> = dt_args
            .get_many::<String>("comment")
            .unwrap()
            .map(String::as_str)
            .flat_map(str::split_whitespace)
            .collect();
        let comment = i2api_utils::parsing::parse_comment(&comment);
        let comment = icinga_api::utils::NonEmptyStr::try_from(comment).unwrap();

        let target = parse_target(dt_args);

        let filter = icinga_api::utils::NonEmptyStr::try_from(target.filter).unwrap();

        let author = icinga_api::utils::NonEmptyStr::try_from(author).unwrap();

        let action = icinga_api::api::ScheduleDowntimeActionBuilder::new(
            author,
            comment,
            end_time,
            filter,
            target.a_type.try_into().unwrap(),
        )
        .unwrap()
        .filter_vars(target.filter_vars.into())
        .start_time(start_time)
        .unwrap()
        .build();
        Arc::new(action)
    };

    rt.block_on(join_all(icingas.map(|service| {
        let action = action.clone();
        rt.spawn(async move {
            service.schedule_downtime(&action).await.unwrap();
        })
    })));
    Ok(())
}

fn rm<Connector>(
    rt: &tokio::runtime::Runtime,
    author: &str,
    icingas: impl Iterator<Item = Arc<icinga::Service<Connector>>>,
    rm_args: &clap::ArgMatches,
) where
    Connector: Connect + Clone + Send + Sync + Debug + 'static,
{
    let target = parse_target(rm_args);
    let downtime_type = match target.a_type {
        icinga_api::api::object_type::ObjectType::Service => target.a_type,
        icinga_api::api::object_type::ObjectType::Host => {
            icinga_api::api::object_type::ObjectType::Downtime
        }
        _ => unreachable!(),
    };
    let filter = icinga_api::utils::NonEmptyStr::try_from(target.filter).unwrap();
    let filter_vars = HashMap::from(target.filter_vars);

    let downtime_action = icinga_api::api::RemoveDowntimeActionBuilder::new(
        filter.clone(),
        downtime_type.try_into().unwrap(),
    )
    .author(author)
    .filter_vars(filter_vars.clone())
    .build();
    let downtime_action = Arc::new(downtime_action);

    let acknowledge_action2 = match target.a_type {
        icinga_api::api::object_type::ObjectType::Host => Some(
            icinga_api::api::RemoveAcknowledgementActionBuilder::new(
                filter.clone(),
                icinga_api::api::remove_acknowledgement_action::TargetType::Service,
            )
            .author(author)
            .filter_vars(filter_vars.clone())
            .build(),
        ),
        icinga_api::api::object_type::ObjectType::Service => None,
        _ => unreachable!(),
    };
    let acknowledge_action2 = acknowledge_action2.map(Arc::new);

    let acknowledge_action1 = icinga_api::api::RemoveAcknowledgementActionBuilder::new(
        filter,
        target.a_type.try_into().unwrap(),
    )
    .author(author)
    .filter_vars(filter_vars)
    .build();
    let acknowledge_action1 = Arc::new(acknowledge_action1);

    let handles = icingas.flat_map(|service| {
        let service_dt = service.clone();
        let service_ack1 = service.clone();

        let mut handles = Vec::new();
        if let Some(acknowledge_action2) = acknowledge_action2.clone() {
            let service_ack2 = service;
            handles.push(rt.spawn(async move {
                service_ack2
                    .remove_acknowledgement(&acknowledge_action2)
                    .await
                    .unwrap();
            }));
        }

        let acknowledge_action1 = acknowledge_action1.clone();
        handles.push(rt.spawn(async move {
            service_ack1
                .remove_acknowledgement(&acknowledge_action1)
                .await
                .unwrap();
        }));

        let downtime_action = downtime_action.clone();
        handles.push(rt.spawn(async move {
            service_dt.remove_downtime(&downtime_action).await.unwrap();
        }));
        handles
    });

    rt.block_on(join_all(handles));
}
